/*  apply USGS NEIC methodology */
/* assume that we have done a synchronize o
   This also assimes that we have a Wood Anderson  seismogram

   Modification history:

   14 JAN 2013 - assume T1 corresponds to vmax arrival
                        T2 corresponds to vmin arrival
			External script will set these for real data sets
			tscal uses these to indicate the beginning and end
			of the windowed time series

	       - use the USGS algorithm to get period and amplitude
   12 MAY 2014 -  modified to use the maximum peak-to-peak in the window
		This was previously mblgmag.c
   28 MAR 2015 - actually implement kt1 kt2 for the window to get the maximum
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "sacsubc.h"
#include "csstime.h"

#define MIN(a,b) ( (b) > (a) ? (a):(b) )
#define MAX(a,b) ( (b) < (a) ? (a):(b) )
#define ABS(a  ) ( (a) >  0  ? (a):-(a))
#define SIGN(a ) ( (a) >  0  ? (1):(-1))

#define EW_SUCCESS 1
#define EW_FAILURE 2
#define EW_WARNING 3

/* define function prototypes */
void gcmdln(int argc, char **argv);
void usage(void);
void get_gain(float freq,float nfreq,float *disp_sens,float df,float *gain);
void getml(float vmin, float vmax, float dt, int npts, float dist, float b, float a, float o, float t1, float t2, char *kstnm, char *kcmpnm, float az, float baz, float *seis);
void gnomesort(int n, float ar[], int key[] ) ;
static int findMLAmp(float *seis, int kta,  int kt1, int kt2, int npts, float dt, float maxMLPeriod, float minMLPeriod, float *ppamp, float *ppper);

/* global parameters */
char *seismogram,  *dispsens;
int have_seismogram, have_dispsens;
/* eventually put these in the command line */
float vmin=3.2;
float vmax=3.8;  
float permin=0.02;
float permax=5.0;
float snlevel=3.0 ;
float *seis;
float *disp_sens;
float df;
int nfreq;
int dosn;

int main(int argc, char **argv)
{
	int nerr;
	int npts;
	int maxpts;
	float b, e, a, dt, o, edist, evdp, dist, az, baz;
	float t1, t2;
	float depmax, depmin;
	char kstnm[9], kcmpnm[9];
	int i;
	/* parse command line arguments */
        gcmdln(argc, argv);
	/* read the header of the seismogram so that we can set up the data structure */
	brsach(seismogram,&nerr)
			;
	/* get the number of poiunts in the time series */
	getnhv("NPTS    ",&npts, &nerr) ;
	maxpts = npts;


	/* get the seismogram */
	brsac(maxpts,seismogram,&seis,&nerr);
		if (nerr < 0 )return -1;
	/* get header values  */
	getfhv("DELTA   ",&dt,&nerr) ;
	getfhv("B       ",&b,&nerr) ;
	getfhv("O       ",&o,&nerr) ;
	getfhv("E       ",&e,&nerr) ;
	getfhv("A       ",&a,&nerr) ;
	getfhv("T1      ",&t1,&nerr) ;
	getfhv("T2      ",&t2,&nerr) ;
	getfhv("DIST    ",&edist,&nerr) ; /* in km */
	getfhv("EVDP    ",&evdp,&nerr) ; /* in km */
	getfhv("AZ      ",&az,&nerr) ;
	getfhv("BAZ     ",&baz,&nerr) ;
	getnhv("NPTS    ",&npts, &nerr) ;
	getkhv("KSTNM   ",kstnm,&nerr);
	getkhv("KCMPNM  ",kcmpnm,&nerr);
	/*
	getfhv("DEPMAX",&depmax,&nerr) ;
	getfhv("DEPMIN",&depmin,&nerr) ;
	printf("DEPMAX: %f\n",depmax);
	printf("DEPMIN: %f\n",depmin);
	*/

	/* get hypocentral distance */
        dist = sqrt(edist*edist + evdp*evdp) ;


	/* get the displacement sensitivity */
	brsac(maxpts,dispsens,&disp_sens,&nerr);
	getfhv("DELTA   ",&df,&nerr) ;
	getnhv("NPTS    ",&nfreq, &nerr) ;
	
	/* human information */
	/*
fprintf(stderr,"edist %f evdp%f dist %f\n",edist, evdp, dist);
	printf("seismogram   : %s\n",seismogram); printf("B            : %f\n",b); printf("O            : %f\n",o);
	printf("DELTA        : %f\n",dt); printf("DIST         : %f\n",dist); printf("AZ           : %f\n",az);
	printf("BAZ          : %f\n",baz); printf("NPTS         : %d\n",npts); printf("KSTNM        : %s\n",kstnm);
	printf("KCMPNM       : %s\n",kcmpnm);

	printf("dispsens     : %s\n",dispsens); printf("DF           : %f\n",df); printf("NFREQ        : %d\n",nfreq);

	*/


	getml(vmin,  vmax, dt, npts, dist, b, o, a, t1, t2, kstnm, kcmpnm, az, baz, seis);

	/* clean up */
	free(seis);
	free(disp_sens);


}	

#define MlPreEventSignal 10

void getml(float vmin, float vmax, float dt, int npts, float dist, float b, float o, float a, float t1, float t2, char *kstnm, char *kcmpnm, float az, float baz, float *seis)
{
	/* allocate a data structure for the absolute value and for the arrays of periods and
	   amps */
	float period, amp;
	float gain, freq;
	int kmin, kmax;	/* indices to vmax and vmin points in array */
	int ko, kb ; /* array pointers to o and b */
	int kta, kt1, kt2; /* pointers to t1 and t2 */
	int num_p2p;
	int i,k,j;
	int num_peak;
	float gamma;
	float slope_minus, slope_plus, second_deriv;
	float ML;
	float maxMLPeriod ;
	float minMLPeriod ;
	float ppamp, ppper;

	kb = 0;
	ko = (int)((o-b)/dt);
	kta = (int)((a -b)/dt);
	kt1 = (int)((t1-b)/dt);
	kt2 = (int)((t2-b)/dt);
/*
fprintf(stderr,"kb %d ko %d kta %d kt1 %d kt2 %d\n",kb,ko,kta,kt1,kt2);
fprintf(stderr,"b %f o %f a %f t1 %f t2 %f\n",b,o,a,t1,t2);
*/
	
	maxMLPeriod = permax;
	minMLPeriod = permin;
	/* compute the magnitude */
	if( findMLAmp(seis,  kta, kt1, kt2, npts, dt, maxMLPeriod, minMLPeriod,&ppamp,&ppper) == EW_SUCCESS){
		/* convert meters to nanometers */
		ppamp *= 1.0e+9 ;

		gamma = 0.00063 ;
			freq = 1.0/ppper;
			get_gain(freq,nfreq,disp_sens,df,&gain);
		ML =  log10(0.5*ppamp) + 1.11*log10(dist) + 0.00189*dist -2.09 ;
		printf("%s %s %f %f %f %f %f %f %f %f\n",kstnm, kcmpnm,dist,az,baz,0.5*ppamp/gain,ppper,ML,0.5*ppamp,gain);
		/* WCI      BHZ      730.567627 274.767090 89.645256 27.036499 0.799988 6.242217  */
	} else {
		printf("%s %s -12345. -12345. -12345. 0.0 0.0 -12345. -12345. -12345.\n",kstnm,kcmpnm);
	}
	return ;
}

void get_gain(float freq,float nfreq,float *disp_sens,float df,float *gain)
{
	/* assume everything is in order of increasing frequency */
	int ilow, ihigh;
	float p;
	ilow = (int)(freq/df);
	ihigh = ilow+1;
	p = (freq - ilow*df)/df;
	*gain = (1-p)*disp_sens[ilow] + p*disp_sens[ihigh];
	return;

}

void gcmdln(int argc, char **argv)
{
/* parse command line arguments, e.g.,
           mlmag -f seismogram -g displaement_sensitivity
      where
*/
	have_seismogram = 0;
	have_dispsens = 0;
	dosn = 0 ;

	char *cp;
	while(argc-- > 1 ) {
		if(*argv[1] == '-'){
			cp = argv[1];
			cp++;
			if(strcmp("f",cp) == 0 || strcmp("F",cp)==0){
				argv++;
				argc--;
				seismogram = argv[1];
				have_seismogram++;
			} else if(strcmp("g",cp) == 0 || strcmp("G",cp)==0){
				argv++;
				argc--;
				dispsens = argv[1];
				have_dispsens++;
			} else if(strcmp("sn",cp) == 0 ) {
					dosn = 1;
			} else if(strcmp("h",cp) == 0 || strcmp("?",cp)==0){
				usage();
			}
			argv++;
		}
	}
	/* put in some syntax checks */
	/* if no other plots, a sacfile requires a response amplitude file */
	/* a sacfile requires a response amplitude file */
		if(have_seismogram == 0 && have_dispsens > 0) usage();
	/* a response amplitude file requires a sacfile */
		if(have_dispsens == 0 && have_seismogram > 0) usage();
}



void usage(void)
{
	fprintf(stderr,
	"Usage: mlmag -f seismogram -g displacement_sensitivity\n");
	fprintf(stderr," -f seismogram (default none) sac binary seismgram\n");
	fprintf(stderr," -g dist_sens  (default none) sac am file\n");
	fprintf(stderr," -sn   (default none) do SN test\n");
	fprintf(stderr," -h           (default false) online help\n");
        exit(EXIT_SUCCESS);
}

/* Gnome Sort - The Simplest Sort Algorithm
 * http://www.cs.vu.nl/~dick/gnomesort.html
 * */
void gnomesort(int n, float ar[], int key[]) {
	int i;
	float tmp;
	int   kmp;
	i = 1;
	while (i < n) {
		if( i == 0){
		       i++;
		} else if( ar[i-1] <= ar[i]) {
			i++;
		}
		else {
			tmp = ar[i]; ar[i] = ar[i-1]; ar[i-1] = tmp;
			kmp = key[i]; key[i] = key[i-1]; key[i-1] = kmp;
			i--;
		}
	}
}

/*  Routines from USGS - modified to the trace here */
/******************************************************************
 * findMLAmp                                                    *
 *                                                                *
 * given a time series traceData and sampleInterval (in secs),    *
 * finds the peak-to-peak amplitude (in counts), zero-to-peak     *
 * amplitude (in counts), period (in seconds), and the time of    *
 * the maximum amplitude (in epoch seconds)                       *
 *                                                                *
 * original FORTRAN [pkamp] by Ray Buland                         *
 * FORTRAN modified for IOC on 8 April 2004 by Ray Buland         *
 ******************************************************************/
static int findMLAmp(float *seis, int kta,  int kt1, int kt2, int npts, float dt, float maxMLPeriod, float minMLPeriod, float *ppamp, float *ppper)
{
    typedef struct _AmpPeriod
    {
        double z2p;                  /* peak-to-peak max amplitude (in nm)              */
        double p2p;                  /* peak-to-peak max amplitude                      */
        double amp;                  /* largest amplitude                               */
        double period;               /* period for ML (in secs)                       */
        int    mest;                 /* sample index for ML                           */
    } AmpPeriod;

    AmpPeriod   ap[3];               /* three largest amps & associated info            */
    double      lag1sec;             /* Buland's d1                                     */
    long        navetm;              /* filter averaging start (sample value index)     */
    long        arrTimeIndex = 0;    /* index of sample point corresponding to
                                      * arrival time                                    */
    long        LgWinStartIndex = 0; /* index of sample point at zero crossing in data 
                                      * immediately prior to Lg window start            */
    long        LgWinEndIndex = 0;   /* index of sample point at last zero crossing in 
                                      * data prior to Lg window end                     */


    double      sum;                 /* sum of background signal                        */
    double      initMean = 0;        /* average of background signal                    */

    double      lastValue = 0.0;     /* previous demeaned, filtered data value          */

    int         zeroCrossing1;       /* sample index for zeroCrossing (Buland's ntp1)   */
    int         zeroCrossing2;       /* sample index for zeroCrossing (Buland's ntp2)   */
    int         zeroCrossing3;       /* sample index for zeroCrossing (Buland's ntp3)   */
    double      currAmp = 0.0;       /* current demeaned and filtered amplitude
                                      * (Buland's af3)                                  */
    double      lastAmp = 0.0;       /* previous demeaned and filtered amplitude 
                                      * (Buland's af2)                                  */
    double      peakAmpCurr = 0.0;   /* absolute value of signal half max               */
    double      peakAmpPrev = 0.0;   /* absolute value of signal half max - 
                                      * previous value                                  */
    long        i;                   /* loop counter                                    */     
    double      cycleDuration = 0.0; /* time for last complete cycle of data            */
    double      halfpeak2peak = 0.0; /* half peak-to-peak amplitude                     */
    int         rc;                  /* return code                                     */

    double  noiseLevel       = 0.0;  /* Noise level                             */
    double  signalLevel      = 0.0;  /* Signal level                            */
    double SNratio ;

    int endNoiseIndex;
    int startNoiseIndex ;

/*
printf("kt1   : %d\n",kt1);
printf("kt2   : %d\n",kt2);
printf("npts  : %d\n",npts);
printf("dt    : %f\n",dt);
*/
    /* safety */
    if(kt1 >= npts || kt1 < 0 || kt2 >= npts  || kta < 0){
	fprintf(stderr,"kt1 >= npts || kt1 < 0 || kt2 >= npts  || kta < 0 %d %d %d %d\n",kt1,npts,kt2,kta);
	    return EW_WARNING;
	}
    /* get indices for the first zero crossing */
    LgWinStartIndex = kt1 ;
    /* if data point at the Lg window start time does not represent a zero crossing of the data, 
     * go back in time until a zero crossing is found */
    for (i = LgWinStartIndex; i > 0; i--)
    {
        if (SIGN(seis[i]) * SIGN(seis[i-1]) <= 0)
        {
            if (seis[i-1] == 0)
                LgWinStartIndex = i-1;
            else
                LgWinStartIndex = i;
           
            break;
        }
    }

    /* backed up to start of data and still didn't find a zero crossing */
    if (LgWinStartIndex == 0 && SIGN(seis[i]) * SIGN(seis[i+1]) > 0)
    {
        /* go forward from expected Lg start time looking for a zero crossing */

        for (i = LgWinStartIndex; i < kt2 - 1; i++)
        {      
            if (seis[i] * seis[i+1] <= 0)
            {
                if (seis[i] == 0)
                    LgWinStartIndex = i;
                else
                    LgWinStartIndex = i+1;
           
                break;
            }
        }

        /* make sure start index is valid */
        if (LgWinStartIndex > kt2 - 2)
        {
	fprintf(stderr,"LgWinStartIndex > kt2 - 2 %d %d\n",LgWinStartIndex,LgWinStartIndex,kt2);
            return EW_WARNING;
        }
    }
    /* find end index of mbLg search window */
    for (i = kt2- 1; i > LgWinStartIndex; i--)
    {
        if (SIGN(seis[i]) * SIGN(seis[i-1]) <= 0)
        {
            if (seis[i-1] == 0)
                LgWinEndIndex = i-1;
            else
                LgWinEndIndex = i;
           
            break;
        }
    }
	if (LgWinEndIndex == LgWinStartIndex + 1)
        {
		fprintf(stderr,"LgWinEndIndex == LgWinStartIndex + 1\n");
            return EW_WARNING;
        }

    /* do noise check */
    if(dosn){
            endNoiseIndex  = kta -1.0/dt;
	    startNoiseIndex = kta -10.0/dt;
	    if(endNoiseIndex < 0 || startNoiseIndex < 0 ){
		fprintf(stderr,"endNoiseIndex < 0 || startNoiseIndex < 0 %d %d\n",endNoiseIndex,startNoiseIndex );
		    return EW_WARNING;
		}
	    for (i = startNoiseIndex; i < endNoiseIndex; i++)
		            noiseLevel += seis[i] * seis[i];
	    for (i = LgWinStartIndex; i < LgWinEndIndex; i++)
		            signalLevel += seis[i] * seis[i];
	    /* check for dead channel */
	        if (signalLevel == 0.00) {
		fprintf(stderr,"signalLevel == 0.00\n");
	            return EW_WARNING;
	        }
	    if(noiseLevel > 0.0){
		    noiseLevel /= (double)(endNoiseIndex - startNoiseIndex);
		    signalLevel /= (double)(LgWinEndIndex - LgWinStartIndex + 1);
		    SNratio = signalLevel / noiseLevel;
		    if(SNratio < snlevel) {
			fprintf(stderr,"SNratio < snlevel\n");
			return EW_WARNING;
			}
            }
    }
    /* initialize zero crossing indices to index of zero crossing preceding
     * Lg window start time */
    zeroCrossing1 = LgWinStartIndex;
    zeroCrossing2 = LgWinStartIndex;
    zeroCrossing3 = LgWinStartIndex;

    /* initialize AmpPeriod structures */
    for (i = 0; i < 3; i++)
        memset(&ap[i], 0, sizeof(AmpPeriod)); 

    /* Loop through the data, measuring three largest peak amplituds  and 
     * corresponding periods between zero crossings */
    /*
printf("LgWinStartIndex %d LgWinEndIndex %d\n",LgWinStartIndex,LgWinEndIndex);
    */
    for (i = LgWinStartIndex; i < LgWinEndIndex; i++)
    {
        lastAmp = currAmp;
        currAmp = seis[i];

        /* collect peak amplitude value */
        peakAmpCurr = MAX(peakAmpCurr, fabs(lastAmp));

        /* Look for a zero crossing */
        if ((SIGN(lastAmp) * SIGN(currAmp) <= 0.0) && (lastAmp != 0.0))
        {
            /* update most recent zero crossing */
            zeroCrossing1 = i;

            /* time for complete cycle */
            cycleDuration = (double)(zeroCrossing1 - zeroCrossing3) * dt;

            if (zeroCrossing2 > LgWinStartIndex &&
                cycleDuration <= maxMLPeriod && cycleDuration >= minMLPeriod)
            {
                halfpeak2peak = (peakAmpCurr + peakAmpPrev) / 2.0;
               
                /* Check to see if this amplitude is a keeper */
                if (halfpeak2peak > ap[2].amp)
                {
                    /* if new amp is larger than all three saved amps, put it in slot 1 */
                    if (halfpeak2peak > ap[0].amp)
                    {
                        memcpy(&ap[2], &ap[1], sizeof(AmpPeriod));
                        memcpy(&ap[1], &ap[0], sizeof(AmpPeriod));
                        ap[0].p2p     = peakAmpCurr + peakAmpPrev;
                        ap[0].z2p     = MAX(peakAmpCurr, peakAmpPrev);
                        ap[0].amp     = halfpeak2peak;
                        ap[0].period  = cycleDuration;
                        ap[0].mest    = zeroCrossing2;
                    }
                    /* if larger than second largest amp, but smaller than
                     * largest amp, put it in slot 2 */
                    else if (halfpeak2peak > ap[1].amp)
                    {
                        memcpy(&ap[2], &ap[1], sizeof(AmpPeriod));
                        ap[1].p2p     = peakAmpCurr + peakAmpPrev;
                        ap[1].z2p     = MAX(peakAmpCurr, peakAmpPrev);
                        ap[1].amp     = halfpeak2peak;
                        ap[1].period  = cycleDuration;
                        ap[1].mest    = zeroCrossing2;
                    }
                    /* largest than third largest amp, but smaller than 
                     * second largest amp */
                    else
                    {
                        ap[2].p2p     = peakAmpCurr + peakAmpPrev;
                        ap[2].z2p     = MAX(peakAmpCurr, peakAmpPrev);
                        ap[2].amp     = halfpeak2peak;
                        ap[2].period  = cycleDuration;
                        ap[2].mest    = zeroCrossing2;
                    }
                }
            }

            /* Update the place holders */
            zeroCrossing3 = zeroCrossing2;
            zeroCrossing2 = zeroCrossing1;
            peakAmpPrev = peakAmpCurr;
            peakAmpCurr = 0;

        } /* if zero crossing */
    } /* loop over data */
    /*
    for(i=0;i<3;i++){
    	printf("ap[%d].p2p:    %f\n",i,ap[i].p2p);
    	printf("ap[%d].z2p:    %f\n",i,ap[i].z2p);
    	printf("ap[%d].amp:    %f\n",i,ap[i].amp);
    	printf("ap[%d].period: %f\n",i,ap[i].period);
    	printf("ap[%d].mest:   %d\n",i,ap[i].mest);
    }
    */

    /* will return ap[0].amp and ap[0].period */
    /*
    printf("ap[0].p2p:    %f\n",ap[0].p2p);
    printf("ap[0].z2p:    %f\n",ap[0].z2p);
    printf("ap[0].amp:    %f\n",ap[0].amp);
    printf("ap[0].period: %f\n",ap[0].period);
    printf("ap[0].mest:   %d\n",ap[0].mest);
    */
    *ppamp = ap[0].p2p;
    *ppper = ap[0].period;
    if( ap[2].p2p == 0){
	fprintf(stderr,"ap[2].p2p == 0\n");
	    return  EW_WARNING;
    } else
    	return EW_SUCCESS;


}
